/*
   This file is part of a program called sphereEversion.
   The complete source code can be downloaded from
      http://www.dgp.toronto.edu/~mjmcguff/eversion/

   This file was written by Michael McGuffin, 1998,
   and has undergone slight modifications since then.
*/
#include <stdio.h>    /* sprintf(), printf(), fprintf(), getchar() */
#include <string.h>   /* memcpy(), strcat() */
#include <stdlib.h>   /* exit() on Linux */
#include <unistd.h>   /* exit() on Irix */
#include <libgen.h>   /* basename() */
#include <math.h>     /* acos() */

/* #include <X11/X.h> */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
/* #include <X11/Xutil.h> */
#define XK_LATIN1
#include <X11/keysym.h>

#include "vroot.h"     /* very important! */

#include <GL/gl.h>
#include <GL/glx.h>   /* glX*() routines */

#include "generateGeometry.h"

/* #define TOGGLEABLE_WINDOW_DECORATIONS */  /* requires Motif libraries */
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
#include <Xm/Xm.h>
#include <Xm/MwmUtil.h>
#else
/* typedef enum { False = 0, True = 1 } Boolean; */
typedef unsigned char Boolean;
#endif

#define DATAWIDTH 320

typedef unsigned long Pixel;

const double PI = 3.14159265358979323846264338327950;

const float objectRadius = 1.5;
const float objectDistanceFromCamera = 1.0;

const double speedOfAnimatedRotation = 0.01; // must be between 0.0 and 1.0

// ==========================================================================

// global variables

double deltaTime = 1.0/512;

struct XStuff {
   char * commandLineName;

   // Recall that in X terminology,
   // one *display* can consist of
   // many *screens*.
   //
   Display *display;
   Atom wm_delete_window, wm_protocols;
   int screen_num;
   Screen *screen_ptr;
   int screenWidth, screenHeight;      // screen dimensions in pixels
   int screenWidthMM, screenHeightMM;  // screen dimensions in millimeters
   unsigned int screenDepth;
   Window rootWindow;
   Window mainwindow, window;
   unsigned int windowMainWidth, windowWidth, windowHeight;
                // main window and subwindow dimensions in pixels
   unsigned int minOfWidthAndHeight() {
      return windowWidth < windowHeight ? windowWidth : windowHeight;
   }
   Cursor cursor;
   Pixel blackPixel, whitePixel, fundPixel[3];
   GC gc;   /* Graphics context. */
   XFontStruct *Font;
   Colormap colorMap;
} XStuff;

typedef struct ButtonData {
   int x1, x2, y1, y2;
   int label;
   KeySym keycode;
} ButtonData;

ButtonData buttonData[30];

// display parameters
const int defaultNumberOfLatitudinalPatchesPerHemisphere = 48;
const int defaultNumberOfLongitudinalPatchesPerStrip = 48;
const int defaultNumStrips = 8;
int colorChannel = 0;
Boolean bBackFaceCulling = False;
Boolean bFrontFacing = True;
Boolean bShadingIsSmooth = True;
typedef enum {
   style_points = 0,
   style_bands,
   style_checkered,
   style_polygons,
   number_of_styles
} RenderingStyle;
RenderingStyle renderingStyle = style_polygons;
Boolean bHalfStrips = False;
Boolean bPause = True;
Boolean bAnimatedRotation = True;
Boolean bAnimatedEversion = True;
Boolean bAnimatedEversionBackwards = False;
Boolean bDisplayText = True;
Boolean bCmdMode = False;
Boolean bRootMode = False;
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
Boolean bNoWindowDecorations = False;
#endif
double angleOfMouseDragForAnimatedRotation = 0.0; // radians
GLfloat rgbdata[24];
Pixel textPixel, panelPixel;

void Help()
{
   int i, charHeight;
   char * help[] = {
      "This program displays a sphere undergoing the Thurston eversion.",
      "To rotate, click and drag with the left mouse button.",
      "",
      "Keys:",
      "  keypad  +  -       Increase/decrease t by delta_t",
      "  keypad  *  /       Multiply/divide delta_t by 2",
      "  page up/down       Increase/decrease total number of strips",
      "  right/left arrows  Increase/decrease number of strips displayed",
      "  up/down arrows     Zoom in/out",
      "  home               Reset zoom to initial value",
      "  end                Toggle display of lower hemisphere",
      "  delete             Toggle display of half-strips",
      "  tab &              Cycle through polygons/checkered polygons/bands/points",
      "  space              Pause / restart animation",
      "  F1                 Toggle animated eversion",
      "  F2                 Toggle animated rotation",
      "  F3/F4              Increase/decrease number of latitudinal patches",
      "  F5/F6              Increase/decrease number of longitudinal patches",
      "  F7                 Reset number of patches to default (48)",
      "  F8                 Toggle command panel",
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
      "  F10                Toggle window decorations",
      "                     (requires Motif-like window manager)",
#endif
      "  s                  Toggle smooth/flat shading",
      "  b                  Toggle backface culling",
      "  f                  Toggle which faces are front facing",
      "  r                  Cycle through rotation modes",
      "  c                  Cycle through color channels",
      "                         front face & back face (/)",
      "                         front face (0), back face (1),",
      "                         background (2), text(3), panel()",
      "  1-8                Select predefined color",
      "  0                  Darker color",
      "  9                  Lighter color",
      "",
      "Press Space to continue ... ",
      NULL };

      XFlush(XStuff.display);     
      usleep(20000);
      if (NULL != XStuff.Font) {
         XSetFont(XStuff.display, XStuff.gc, XStuff.Font->fid);
         charHeight = (int)(
            1.2 * ( XStuff.Font->ascent + XStuff.Font->descent )
         );
      }

      XSetForeground(XStuff.display, XStuff.gc, XStuff.blackPixel);
      XFillRectangle(XStuff.display, XStuff.window, XStuff.gc, 0, 0,
		     XStuff.windowMainWidth, XStuff.windowHeight);
      XSetForeground(XStuff.display, XStuff.gc, textPixel);

      for (i=0; help[i]; i++)
         XDrawString(XStuff.display, XStuff.window, XStuff.gc, 
                     10, charHeight*i + 25, help[i], strlen(help[i]));
      XFlush(XStuff.display);
}

// ==========================================================================

Pixel AllocPixelColor(double r, double g, double b) {

   XColor xcolor;

   xcolor.pixel = 0;
   xcolor.red   = (int)(r * 65535 + 0.5);  // add .5 to round-off to nearest integer
   xcolor.green = (int)(g * 65535 + 0.5);
   xcolor.blue  = (int)(b * 65535 + 0.5);
   xcolor.flags = DoRed | DoGreen | DoBlue;
   XAllocColor(XStuff.display, XStuff.colorMap, &xcolor);
   return xcolor.pixel;
}

void FreePixelColor(Pixel p) {

   XFreeColors(XStuff.display, XStuff.colorMap, &p, 1, 0);
}

// ==========================================================================

void MakeWindow(
   char * windowTitle,
   char * iconTitle,
   Pixmap iconPixmap,                      /* Pass (None) for none. */
   int minWidth, int minHeight,
   Boolean isOpenGLWindow,
   Pixel backgroundPixel,
   int argc, char *argv[]                  /* Command line arguments. */
) {
   XWMHints *wm_hints;                     /* Window manager hints. */
   XClassHint *class_hints;                /* Application class hints. */
   XSizeHints *size_hints;                 /* Preferred window geometry. */
   XColor xc;
   XTextProperty windowName,iconName;      /* Special X "strings" */
   int windowX, windowY;                   /* Upper left corner of window. */
   unsigned int windowBorderWidth;

   /* This stuff is for creating an OpenGL window. */
   GLXContext glxContext;
   XVisualInfo * visualInfo;
   XSetWindowAttributes setableAttributes;

   /*
      Allocate hint structures.
   */
   if (NULL == (wm_hints = XAllocWMHints()) ||
      NULL == (class_hints = XAllocClassHint()) ||
      NULL == (size_hints = XAllocSizeHints())
   ) {
      fprintf(stderr,"%s: failure allocating memory.\n", XStuff.commandLineName);
      exit(1);
   }

   /*
      Create text properties.
   */
   if (0 == XStringListToTextProperty(&windowTitle, 1, &windowName) ||
      0 == XStringListToTextProperty(&iconTitle, 1, &iconName)
   ) {
      fprintf(stderr,"%s: structure allocation for text property failed.\n",
         XStuff.commandLineName);
      exit(1);
   }

   if (isOpenGLWindow) {
      int attributeList[] = {
         GLX_RGBA,
         GLX_DOUBLEBUFFER,
         GLX_RED_SIZE, 1,
         GLX_GREEN_SIZE, 1,
         GLX_BLUE_SIZE, 1,
         GLX_DEPTH_SIZE, 1,
         None
      };

      /* Open an X visual */
      if ( bRootMode ) {

        XWindowAttributes xgwa;
        XVisualInfo vi_in, *vi_out;
        int out_count;

        XStuff.window = XStuff.mainwindow = XStuff.rootWindow;
        XGetWindowAttributes (XStuff.display, XStuff.window, &xgwa);

        vi_in.screen = XStuff.screen_num;
        vi_in.visualid = XVisualIDFromVisual (xgwa.visual);
        vi_out = XGetVisualInfo (XStuff.display, VisualScreenMask|VisualIDMask,
                                 &vi_in, &out_count);
        if (!vi_out) {
          fprintf(stderr,"%s: can't get GL visual for window 0x%lx.\n",
                  XStuff.commandLineName, (unsigned long) XStuff.window);
          exit(1);
        }

        visualInfo = vi_out;
      } else {
        if (! (visualInfo = glXChooseVisual(XStuff.display, XStuff.screen_num, attributeList))) {
          fprintf(stderr,"%s: can't open GL visual.\n", XStuff.commandLineName);
          exit(1);
        }
      }

      /* Create a GLX context */
      if (!(glxContext = glXCreateContext(XStuff.display, visualInfo, None, GL_TRUE))) {
         fprintf(stderr,"%s: can't open GLX context.\n", XStuff.commandLineName);
         exit(1);
      }

      /* create a colormap */
      XStuff.colorMap = XCreateColormap(XStuff.display, XStuff.rootWindow, visualInfo->visual, AllocNone);
   }

   /*
      Create the main window.
   */
   windowX = XStuff.screenWidth/20;
   windowY = XStuff.screenHeight/20;
   XStuff.windowMainWidth = XStuff.screenWidth*9/10;
   XStuff.windowWidth = XStuff.windowMainWidth - DATAWIDTH;
   XStuff.windowHeight = XStuff.screenHeight*9/10;
   windowBorderWidth = 0;

   if (isOpenGLWindow) {
      setableAttributes.colormap = XStuff.colorMap;
      setableAttributes.border_pixel = 0;

      if ( bRootMode ) {
         XStuff.window = XStuff.rootWindow;
         XStuff.mainwindow = XCreateWindow(
            XStuff.display, XStuff.rootWindow,
            windowX, windowY, 
            DATAWIDTH, 600,
            windowBorderWidth,
            visualInfo->depth, InputOutput, visualInfo->visual,
            CWBorderPixel | CWColormap,
            &setableAttributes);
      }
      else {
         XStuff.mainwindow = XCreateWindow(
            XStuff.display, XStuff.rootWindow,
            windowX, windowY, XStuff.windowMainWidth, XStuff.windowHeight,
            windowBorderWidth,
            visualInfo->depth, InputOutput, visualInfo->visual,
            CWBorderPixel | CWColormap,
            &setableAttributes);
         XStuff.window = XCreateWindow(
            XStuff.display, XStuff.mainwindow,
            0, 0, XStuff.windowWidth, XStuff.windowHeight,
            windowBorderWidth,
            visualInfo->depth, InputOutput, visualInfo->visual,
            CWBorderPixel | CWColormap,
            &setableAttributes);
      }
   }
   else {
      XStuff.mainwindow = XCreateSimpleWindow(XStuff.display, XStuff.rootWindow,
         windowX, windowY, XStuff.windowMainWidth, XStuff.windowHeight,
         windowBorderWidth,
         XStuff.blackPixel, XStuff.whitePixel);
      XStuff.window = XCreateSimpleWindow(XStuff.display, XStuff.mainwindow,
         0, 0, XStuff.windowWidth, XStuff.windowHeight,
         windowBorderWidth,
         XStuff.blackPixel, XStuff.whitePixel);
   }

   /*
      Set window properties.
   */
   wm_hints->flags = StateHint | InputHint;
   wm_hints->initial_state = NormalState;   /* Should window be normal or iconified when first mapped ? */
   wm_hints->input = True;                  /* Does application need keyboard input? */
   if (None != iconPixmap) {
      wm_hints->flags |= IconPixmapHint;
      wm_hints->icon_pixmap = iconPixmap;
   }

   /* These are used by the window manager to get information */
   /* about this application from the resource database. */
   /*
      For example, with 4Dwm, you could add lines
      like these to your ~/.Xdefaults file :
              4Dwm*foo.clientDecoration: none
              4Dwm*goo.clientDecoration: border resizeh
      If foo is the name or class of an app's window, then that window will
      have no window decorations.  Similarly, in the case of goo, the window
      will only have a resizable border.  (The clientDecoration resource works
      the same way with Mwm.)
   */
   class_hints->res_name = XStuff.commandLineName;
   class_hints->res_class = "MyClassOfApplications";

   /* Use USPosition | USSize instead of PPosition | PSize to force hints. */
   /* size_hints->flags = PPosition | PSize | PMinSize; */
   size_hints->flags = PMinSize;
   /* In R4 and later, the x, y, width, and height members of XSizeHints should not be set. */
   size_hints->min_width = minWidth;
   size_hints->min_height = minHeight;

   XSetWMProperties(XStuff.display, XStuff.mainwindow, &windowName, &iconName, 
        argv, argc, size_hints, wm_hints,class_hints);
   XSetWMProtocols(XStuff.display, XStuff.mainwindow, &XStuff.wm_delete_window, 1);

   /*
      Select event types wanted.

      ConfigureNotify events inform the app that the window has been
      resized.  Processing ConfigureNotify events is more efficient
      than having to call XGetGeometry() (which requires a reply from
      the server) on every Expose event.
   */

   XSelectInput(XStuff.display, XStuff.mainwindow,
      KeyPressMask |
      KeyReleaseMask |
      ButtonPressMask |
      ButtonReleaseMask |
      StructureNotifyMask /* selects CirculateNotify, ConfigureNotify, DestroyNotify, GravityNotify, MapNotify, ReparentNotify, and UnmapNotify */
      );

   if (!bRootMode)
   XSelectInput(XStuff.display, XStuff.window,
      ExposureMask |
      KeyPressMask |
      KeyReleaseMask |
      ButtonPressMask |
      ButtonReleaseMask |
      PointerMotionMask |
      PointerMotionHintMask
      );
   /*
      Display the window.
   */

   if (bRootMode ) {
      if (bCmdMode) {
         XMapRaised(XStuff.display, XStuff.mainwindow);
         XFlush(XStuff.display);
         usleep(40000);
         XSetInputFocus(XStuff.display, XStuff.mainwindow, RevertToNone, CurrentTime);
      }
   } else {
      XMapRaised(XStuff.display, XStuff.mainwindow);
      XMapRaised(XStuff.display, XStuff.window);
      XFlush(XStuff.display);
   }

   if (isOpenGLWindow) {
      /* Connect the context to the window. */
      if (!glXMakeCurrent(XStuff.display, XStuff.window, glxContext)) {
         fprintf(stderr,"%s: can't set GL context.\n", XStuff.commandLineName);
         exit(1);
      }
   }

   if (!isOpenGLWindow) {
      /*
         Set the background color.
      */
      XSetWindowBackground(XStuff.display, XStuff.window, backgroundPixel);
   }

   // These fonts will be used if the
   // former ones cannot be found.
   //
   char * nameOfFont = "9x15";

   if (NULL == (XStuff.Font = XLoadQueryFont(XStuff.display,nameOfFont)))
         fprintf(stderr,"%s: can't open font \"%s\"\nSome text may not appear.\n", 
            XStuff.commandLineName, nameOfFont
         );

   /*
      Get a graphics context.
   */
   XStuff.gc = XCreateGC(XStuff.display, XStuff.window, 0, NULL);
   /* Specify black foreground since default window background */
   /* is white and default foreground is undefined. */
   XSetForeground(XStuff.display, XStuff.gc, XStuff.blackPixel);

   /*
      Get geometry information about window
   */
   if (False == XGetGeometry(XStuff.display, XStuff.window, &XStuff.rootWindow,
      &windowX, &windowY, &XStuff.windowWidth, &XStuff.windowHeight,
      &windowBorderWidth, &XStuff.screenDepth)
   ) {
      fprintf(stderr,"%s: can't get window geometry.\n", XStuff.commandLineName);
      exit(1);
   }
   XStuff.windowMainWidth = XStuff.windowMainWidth + DATAWIDTH;
}

// ==============================================================================

// The below bitmaps were created with the command "bitmap [filename]".

#define cursor_width 25
#define cursor_height 25
#define cursor_x_hot 12
#define cursor_y_hot 12
static char cursor_bits[] = {
   0xff, 0xff, 0xff, 0xff, 0xdf, 0xff, 0xf7, 0xff, 0xdf, 0xff, 0xf7, 0xff,
   0xaf, 0xff, 0xeb, 0xff, 0xaf, 0xff, 0xeb, 0xff, 0x77, 0xff, 0xdd, 0xff,
   0x77, 0xff, 0xdd, 0xff, 0xfb, 0xfe, 0xbe, 0xff, 0xfb, 0xfe, 0xbe, 0xff,
   0x01, 0x7c, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff,
   0xff, 0xd7, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff, 0xff, 0xef, 0xff, 0xff,
   0xff, 0xd7, 0xff, 0xff, 0xff, 0xd7, 0xff, 0xff, 0xff, 0xbb, 0xff, 0xff,
   0xff, 0xbb, 0xff, 0xff, 0xff, 0x7d, 0xff, 0xff, 0xff, 0x7d, 0xff, 0xff,
   0xff, 0x00, 0xfe, 0xff};

#define mask_width 25
#define mask_height 25
#define mask_x_hot 12
#define mask_y_hot 12
static char mask_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x08, 0x00, 0x20, 0x00, 0x08, 0x00,
   0x70, 0x00, 0x1c, 0x00, 0x70, 0x00, 0x1c, 0x00, 0xf8, 0x00, 0x3e, 0x00,
   0xf8, 0x00, 0x3e, 0x00, 0xfc, 0x01, 0x7f, 0x00, 0xfc, 0x01, 0x7f, 0x00,
   0xfe, 0x83, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00,
   0x00, 0x38, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00,
   0x00, 0x38, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00, 0x7c, 0x00, 0x00,
   0x00, 0x7c, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00, 0x00, 0xfe, 0x00, 0x00,
   0x00, 0xff, 0x01, 0x00};

void CreateCursorForWindow() {

   Pixmap source, mask;
   XColor foreground, background;

   foreground.pixel = 0;
   foreground.red = 0;
   foreground.green = 65535;
   foreground.blue = 0;
   foreground.flags = DoRed | DoGreen | DoBlue;
   foreground.pad = 0;

   background.pixel = 0;
   background.red = 0;
   background.green = 0;
   background.blue = 0;
   background.flags = DoRed | DoGreen | DoBlue;
   background.pad = 0;

   source = XCreateBitmapFromData(XStuff.display, XStuff.window,
                   cursor_bits, cursor_width, cursor_height);
   mask = XCreateBitmapFromData(XStuff.display, XStuff.window,
                   mask_bits, mask_width, mask_height);
   XStuff.cursor = XCreatePixmapCursor(XStuff.display, source, mask,
                   &foreground, &background,
                   cursor_x_hot, cursor_y_hot);

   XDefineCursor(XStuff.display, XStuff.window, XStuff.cursor);

   XFreePixmap(XStuff.display, source);
   XFreePixmap(XStuff.display, mask);
}

// ==============================================================================

class Frustum {
public:
   void Init();
   void WindowResize();
   void Magnify(int delta);  // delta can be positive (to zoom in) or negative (to zoom out)
   void Reset();

   double Width() { return width; }
   double Height() { return height; }
   double LeftEdge() { return center_x - width/2; }
   double TopEdge() { return center_y + height/2; }
   double NearPlane() { return near; }
   // magnificationFactor == magIncrement ^ magExponent
   //
   double magnificationFactor;  // 1.0 by default, larger values shrink the (width,height) proportionally

private:
   double unMagnifiedSize;
   double size;
   double width;   // width of near plane
   double height;  // height of near plane
   double near;    // distance to near plane from observer
   double far;     // distance to far plane from observer
   double center_x;
   double center_y;

   int magExponent;             // 0 by default
   static const double magIncrement;
};

const double Frustum::magIncrement = 1.05;

void Frustum::Init() {

   near = objectDistanceFromCamera;
   far = near + 4*objectRadius;
   center_x = 0.0;
   center_y = 0.0;

   magnificationFactor = 1.0;
   magExponent = 0;

   // Create clipping volume.
   // Let r be objectRadius, d be objectDistanceFromCamera.
   // We have two similar right triangles
   //
   //                         |
   //                  |?     |r
   // eye  ____________|______|
   //            d        r
   //
   //
   //  ? = alpha = dr/(r+d)
   //
   // But this value of alpha makes the object unpleasantly
   // large, so we multiply by a constant greater than 1.0
   //
   double alpha = 1.25 * objectDistanceFromCamera * objectRadius / (objectRadius + objectDistanceFromCamera);
   unMagnifiedSize = 2*alpha;
   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

void Frustum::WindowResize() {

   glViewport(0,0, XStuff.windowWidth, XStuff.windowHeight);
   if (XStuff.windowWidth < XStuff.windowHeight) {
      width = size;
      height = size * XStuff.windowHeight/(double)XStuff.windowWidth;
   }
   else {
      height = size;
      width = size * XStuff.windowWidth/(double)XStuff.windowHeight;
   }

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();   // clear matrix
   
   // for orthographic projection, use glOrtho() instead of glFrustum()
   //
   glFrustum(
      center_x - width/2, center_x + width/2, /* left,right */
      center_y - height/2, center_y + height/2, /* bottom, top */
      near, far /* near, far */
   );
}

void Frustum::Magnify(int delta) {

   int j;

   magExponent += delta;
   magnificationFactor = 1.0;
   if (magExponent < 0)
      for (j = magExponent; j < 0; ++j)
         magnificationFactor /= magIncrement;
   else for (j = 0; j < magExponent; ++j)
      magnificationFactor *= magIncrement;

   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

void Frustum::Reset() {

   center_x = 0;
   center_y = 0;
   magExponent = 0;
   magnificationFactor = 1.0;
   size = unMagnifiedSize / magnificationFactor;
   WindowResize();
}

// ==============================================================================

Frustum frustum;

// ==============================================================================

class Sphere {
public:
    Sphere(double time = 0,int numStrips = 0,int numHemispheresToDisplay = 0,int numStripsToDisplay = 0,int numberOfLatitudinalPatchesPerHemisphere = 0,int numberOfLongitudinalPatchesPerStrip = 0) :
        matrixOfVertices(NULL) { Construct(time,numStrips,numHemispheresToDisplay,numStripsToDisplay,numberOfLatitudinalPatchesPerHemisphere,numberOfLongitudinalPatchesPerStrip); }
    ~Sphere();
    void Construct(double time = 0,int numStrips = 0,int numHemispheresToDisplay = 0,int numStripsToDisplay = 0,int numberOfLatitudinalPatchesPerHemisphere = 0,int numberOfLongitudinalPatchesPerStrip = 0);
    void Reconstruct() { DeallocateMatrix(); GenerateVertices(); }
    void Draw();
    void SetTime(double Value) { DeallocateMatrix(); Time = Value; GenerateVertices(); }
    void IncrementTime(double deltaTime) { if (Time < 1.0) { DeallocateMatrix(); Time+=deltaTime; if (deltaTime > 1.0) deltaTime = 1.0; GenerateVertices(); } }
    void DecrementTime(double deltaTime) { if (Time > 0.0) { DeallocateMatrix(); Time-=deltaTime; if (deltaTime < 0.0) deltaTime = 0.0; GenerateVertices(); } }
    void IncrementNumStrips() { DeallocateMatrix(); ++NumStrips; ++NumStripsToDisplay; GenerateVertices(); }
    void DecrementNumStrips() { if (NumStrips > 1) { DeallocateMatrix(); --NumStrips; if (NumStripsToDisplay > 1) --NumStripsToDisplay; GenerateVertices(); } }
    void ChangeNumHemispheresToDisplay() { if (NumHemispheresToDisplay == NumHemispheres) NumHemispheresToDisplay = 1; else ++NumHemispheresToDisplay; }
    void IncrementNumStripsToDisplay() { if (NumStripsToDisplay < NumStrips) {++NumStripsToDisplay;} }
    void DecrementNumStripsToDisplay() { if (NumStripsToDisplay > 1) {--NumStripsToDisplay;} }
    void SetDefaultResolution() { NumberOfLatitudinalPatchesPerHemisphere = NumberOfLongitudinalPatchesPerStrip = 48; GenerateVertices(); }
    void IncrementLatitudinalResolution() { DeallocateMatrix(); ++NumberOfLatitudinalPatchesPerHemisphere; GenerateVertices(); }
    void DecrementLatitudinalResolution() { DeallocateMatrix(); --NumberOfLatitudinalPatchesPerHemisphere; GenerateVertices(); }
    void IncrementLongitudinalResolution() { DeallocateMatrix(); ++NumberOfLongitudinalPatchesPerStrip; GenerateVertices(); }
    void DecrementLongitudinalResolution() { DeallocateMatrix(); --NumberOfLongitudinalPatchesPerStrip; GenerateVertices(); }

    double GetTime() { return Time; }

    int NumStrips;
    int NumStripsToDisplay;
    int NumHemispheresToDisplay;
    int NumberOfLatitudinalPatchesPerHemisphere,
        NumberOfLongitudinalPatchesPerStrip;

private:
    void GenerateVertices();
    void DeallocateMatrix();

    static const int NumHemispheres;
    double Time;   // between 0.0 and 1.0

    // This matrix stores all the vertices and normals used to render the
    // sphere.  Elements in the matrix are arranged by [latitude][longitude].
    //
    GLPoint ** matrixOfVertices;
};

const int Sphere::NumHemispheres = 2;

Sphere::~Sphere() {

    DeallocateMatrix();
}

void Sphere::DeallocateMatrix() {

    int j;
    
    if (matrixOfVertices == NULL)
       return;
    for (j = NumberOfLatitudinalPatchesPerHemisphere; j >= 0; --j)
       delete [] (matrixOfVertices[j]);
    delete [] matrixOfVertices;
    matrixOfVertices = NULL;
}

void Sphere::Construct(
    double time,
    int numStrips,
    int numHemispheresToDisplay,
    int numStripsToDisplay,
    int numberOfLatitudinalPatchesPerHemisphere,
    int numberOfLongitudinalPatchesPerStrip
) {
    if (time < 0.0)
       Time = 0.0;
    else if (time > 1.0)
       Time = 1.0;
    else Time = time;

    NumStrips = numStrips < 1 ? defaultNumStrips : numStrips;

    NumHemispheresToDisplay = numHemispheresToDisplay < 1 ? NumHemispheres : numHemispheresToDisplay;

    NumStripsToDisplay = numStripsToDisplay < 1 ? NumStrips : numStripsToDisplay;

    if (numberOfLatitudinalPatchesPerHemisphere <= 0)
       NumberOfLatitudinalPatchesPerHemisphere = defaultNumberOfLatitudinalPatchesPerHemisphere;
    else
       NumberOfLatitudinalPatchesPerHemisphere = numberOfLatitudinalPatchesPerHemisphere;

    if (numberOfLongitudinalPatchesPerStrip <= 0)
       NumberOfLongitudinalPatchesPerStrip = defaultNumberOfLongitudinalPatchesPerStrip;
    else
       NumberOfLongitudinalPatchesPerStrip = numberOfLongitudinalPatchesPerStrip;

    GenerateVertices();
}

void Sphere::GenerateVertices() {

    int j;

    // clamp input parameters to their minima
    if (Time < 0.0)
       Time = 0.0;
    else if (Time > 1.0)
       Time = 1.0;

    if (NumberOfLatitudinalPatchesPerHemisphere < 2)
       NumberOfLatitudinalPatchesPerHemisphere = 2;

    if (NumberOfLongitudinalPatchesPerStrip < 2)
       NumberOfLongitudinalPatchesPerStrip = 2;

    // throw away previous geometry
    DeallocateMatrix();

    // allocate stuff
    matrixOfVertices = new GLPointPointer[1 + NumberOfLatitudinalPatchesPerHemisphere];
    for (j = NumberOfLatitudinalPatchesPerHemisphere; j >= 0; --j)
       matrixOfVertices[j] = new GLPoint[1 + NumberOfLongitudinalPatchesPerStrip];

    // generate the geometry
    generateGeometry(
       matrixOfVertices,
       Time,
       NumStrips,

       0.0,
       NumberOfLatitudinalPatchesPerHemisphere,
       1.0,
       0.0,
       NumberOfLongitudinalPatchesPerStrip,
       bHalfStrips ? 0.5 : 1.0
#ifdef BEND_IN /* this will display a cylindar bending into a sphere */
       ,Time
#endif
    );

}

void Sphere::Draw() {

    /*
    In a GL_TRIANGLE_STRIP block, if the vertices sent
    in are numbered 0,1,2,3,... the resulting strip will look like

        0---2---4---6---8--- ...
         \ / \ / \ / \ / \
          1---3---5---7---9 ...
 
    and will be cowposed of triangles (0,1,2), (2,1,3), (2,3,4), ...
    The ordering is important.  By convention, polygons whose vertices
    appear in counterclockwise order on the screen are front-facing.
    If you create polygons with vertices in the wrong order, and
    back-face culling is turned on, you'll never see the polygons
    appear on the screen.
    */

    int hemisphere,strip,j,k;

    glFrontFace(GL_CW);  // we're going to use the opposite convention

    glMatrixMode(GL_MODELVIEW);

    for (hemisphere = 0; hemisphere < NumHemispheresToDisplay; ++hemisphere) {
        glPushMatrix();
        glRotatef(hemisphere*180.0,0.0,1.0,0.0);
        for (strip = 0; strip < NumStripsToDisplay; ++strip) {
            glPushMatrix();
            glRotatef((hemisphere == 0 ? -strip : strip+1)*360.0/NumStrips,0.0,0.0,1.0);

            if (renderingStyle == style_points) {
                glBegin(GL_POINTS);
                for (j = 0; j <= NumberOfLatitudinalPatchesPerHemisphere; ++j)
                    for (k = 0; k <= NumberOfLongitudinalPatchesPerStrip; ++k) {
                        glNormal3fv(matrixOfVertices[j][k].normal);
                        glVertex3fv(matrixOfVertices[j][k].vertex);
                    }
                glEnd();
            }
            else {
                for (j = 0; j < NumberOfLatitudinalPatchesPerHemisphere; ++j) {
                    if (renderingStyle == style_polygons || (renderingStyle == style_bands && (j & 1)==hemisphere)) {
                        glBegin(GL_TRIANGLE_STRIP);
                        for (k = 0; k <= NumberOfLongitudinalPatchesPerStrip; ++k) {
                            glNormal3fv(matrixOfVertices[j][k].normal);
                            glVertex3fv(matrixOfVertices[j][k].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k].normal);
                            glVertex3fv(matrixOfVertices[j+1][k].vertex);
                        }
                        glEnd();
                    }
                    else if (renderingStyle == style_checkered) {
                        for (k = j%2; k < NumberOfLongitudinalPatchesPerStrip; k+=2) {
                            glBegin(GL_TRIANGLE_STRIP);
                            glNormal3fv(matrixOfVertices[j][k].normal);
                            glVertex3fv(matrixOfVertices[j][k].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k].normal);
                            glVertex3fv(matrixOfVertices[j+1][k].vertex);
                            glNormal3fv(matrixOfVertices[j][k+1].normal);
                            glVertex3fv(matrixOfVertices[j][k+1].vertex);
                            glNormal3fv(matrixOfVertices[j+1][k+1].normal);
                            glVertex3fv(matrixOfVertices[j+1][k+1].vertex);
                            glEnd();
                        }
                    }
                }
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
}

// ==============================================================================

Sphere sphere;

// ==============================================================================

class RotationInformation {
public:
    RotationInformation()
        : centerX(0), centerY(0), viewportRadius(10),
          isClutchOn(False),
          previousDeviceX(0), previousDeviceY(0),
          rotationMode(arcball),
          elevation(0.0), azimuth(0.0)
    {
        CopyIdentityInto(oldRotationMatrix);
        CopyIdentityInto(rotationMatrix); 
    }

    void CalibrateDevice(int center_x, int center_y, int viewport_radius) {
        centerX = center_x;
        centerY = center_y;
        viewportRadius = viewport_radius;
    }

    Boolean IsClutchOn() { return isClutchOn; }
    void TurnClutchOn(int deviceX, int deviceY) {    // mouse button down
        isClutchOn = True;
        previousDeviceX = deviceX;
        previousDeviceY = deviceY;
    }
    void TurnClutchOff() {   // mouse button up
        isClutchOn = False;
    }
    void DeviceMotion(float deviceX, float deviceY);   // mouse motion

    enum RotationMode {   // CycleRotationMode() assumes that the first mode equals zero
        elevationAzimuth = 0,
        aroundXYAxes,
        virtualSphere,
        arcball,
        numberOfRotationModes
    };
    RotationMode GetRotationMode() { return rotationMode; }
    void SetRotationMode(RotationMode newMode);
    void CycleRotationMode(int shift) { SetRotationMode((RotationMode)(((shift == 1)?(rotationMode+1) % (numberOfRotationModes):(rotationMode+numberOfRotationModes-1) % (numberOfRotationModes)))); }

    float * GetRotationMatrix() { return rotationMatrix; }
private:
    void CopyIdentityInto(float * matrix);

    int centerX, centerY, viewportRadius;
    Boolean isClutchOn;
    int previousDeviceX, previousDeviceY;
    float oldRotationMatrix[16];   // cumulative past rotations
    float rotationMatrix[16];   // oldRotationMatrix + current rotation
    // stuff specific to "elevationAzimuth" mode
    float elevation, azimuth;   // in degrees; not radians
    RotationMode rotationMode;
};

void RotationInformation::DeviceMotion(float deviceX, float deviceY) {

    // Here, the device axis conventions are X+ right, Y+ down
    // Note: glRotatef(theta, x, y, z); rotates theta degrees counter clockwise around vector (x,y,z)

    // The device was dragged deltaX pixels to the right and deltaY pixels down.
    float deltaX = deviceX - previousDeviceX,
          deltaY = deviceY - previousDeviceY;
    float pixelsPerDegree = 0.01*viewportRadius;
    
    if (!isClutchOn)
       return;

    glMatrixMode(GL_MODELVIEW);   // set transformation mode to model/view transformations
    glLoadIdentity();             // clear matrix

    switch (rotationMode) {
    case elevationAzimuth:
        elevation += deltaY/pixelsPerDegree;
        azimuth += deltaX/pixelsPerDegree;
        glRotatef(elevation,1.0,0.0,0.0);   // perform rotation around x+ axis (pointing right)
        glRotatef(azimuth,0.0,1.0,0.0);   // perform rotation around y+ axis (pointing up)
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        // don't save into old
        break;
    case aroundXYAxes:
        glRotatef(deltaY/pixelsPerDegree,1.0,0.0,0.0);   // perform rotation around x+ axis (pointing right)
        glRotatef(deltaX/pixelsPerDegree,0.0,1.0,0.0);   // perform rotation around y+ axis (pointing up)
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        glGetFloatv(GL_MODELVIEW_MATRIX, oldRotationMatrix);   // save result into old
        break;
    case virtualSphere:
    case arcball:
        float ux, uy, uz,   // vector u corresponding to previous mouse position, where |u| = 1
              uzSquared,
              vx, vy, vz,   // vector v corresponding to new mouse position, where |v| = 1
              wx, wy, wz,   // vector w is cross product (u x v)
              vzSquared;
        float theta,        // angle (in radians) between u and v
              dotProduct,
              shrinkFactor;

        // compute u
        //
        ux = (previousDeviceX - centerX) / (float)viewportRadius;
        uy = (centerY - previousDeviceY) / (float)viewportRadius;   // uy+ points up, not down
        uzSquared = 1.0 - ux*ux - uy*uy;
        if (uzSquared < 0.0) {
           // Shorten vector to force a magnitude of one
           shrinkFactor = sqrt(ux*ux + uy*uy);
           ux /= shrinkFactor;
           uy /= shrinkFactor;
           uz = 0.0;
        }
        else uz = sqrt(uzSquared);

        // compute v
        //
        vx = (deviceX - centerX) / (float)viewportRadius;
        vy = (centerY - deviceY) / (float)viewportRadius;   // vy+ points up, not down
        vzSquared = 1.0 - vx*vx - vy*vy;
        if (vzSquared < 0.0) {
           // Shorten vector to force a magnitude of one
           shrinkFactor = sqrt(vx*vx + vy*vy);
           vx /= shrinkFactor;
           vy /= shrinkFactor;
           vz = 0.0;
        }
        else vz = sqrt(vzSquared);

        // compute cross product w = (u x v)
        //
        wx =   uy*vz - vy*uz;
        wy = -(ux*vz - vx*uz);
        wz =   ux*vy - vx*uy;

        // Since
        //    |w| = |u||v||sin theta|
        // and
        //    |u| = |v| = 1
        // we have
        //    |w| = |sin theta|
        // But theta = asin(|w|) only if 0 <= theta <= pi/2 radians.
        // If theta > pi/2, theta != asin(|w|).
        //
        // So we calculate theta using the dot product (u.v)
        //    u.v = |u||v|(cos theta) = cos theta
        //    theta = acos(u.v)
        // This works for 0 <= theta <= pi radians.
        dotProduct = ux*vx + uy*vy + uz*vz;
        if (dotProduct < -1.0)
            dotProduct = -1.0;
        else if (dotProduct > 1.0)
            dotProduct = 1.0;
        theta = acos(dotProduct);
        if (rotationMode == arcball) {
            // This is the only difference between virtualSphere and arcball.
            // Note that this number could be changed to achieve a generalization
            // of virtualSphere/arcball ... although virtualSphere is the only mode
            // which does not exhibit hysterisis.
            theta *= 2.0;
        }
        glRotatef(180 * theta / PI, wx, wy, wz);
        glMultMatrixf(oldRotationMatrix);   // preserve previous rotations
        glGetFloatv(GL_MODELVIEW_MATRIX, oldRotationMatrix);   // save result into old
        break;
      default:
        abort();
        break;
    };
    
    glGetFloatv(GL_MODELVIEW_MATRIX, rotationMatrix);   // save result

    // Forgetting these two lines made for very interesting behaviour.
    // (Some bugs are cool !)  Try commenting them and see what happens ...
    previousDeviceX = (int)deviceX;
    previousDeviceY = (int)deviceY;
}

void RotationInformation::SetRotationMode(RotationMode newMode) {

    // Save rotations of current mode so they won't be lost,
    // allowing us to transfer smoothly into the new mode.
    memcpy(oldRotationMatrix,   // destination
           rotationMatrix,   // source
           16*sizeof(float));

    // reset mode-specific parameters
    elevation = 0.0;
    azimuth = 0.0;
    
    // set to new mode
    rotationMode = newMode;
}

void RotationInformation::CopyIdentityInto(float * matrix) {

    matrix[ 0] = 1.0;   matrix[ 1] = 0.0;   matrix[ 2] = 0.0;   matrix[ 3] = 0.0;
    matrix[ 4] = 0.0;   matrix[ 5] = 1.0;   matrix[ 6] = 0.0;   matrix[ 7] = 0.0;
    matrix[ 8] = 0.0;   matrix[ 9] = 0.0;   matrix[10] = 1.0;   matrix[11] = 0.0;
    matrix[12] = 0.0;   matrix[13] = 0.0;   matrix[14] = 0.0;   matrix[15] = 1.0;
}

// ============================================================================

RotationInformation Rotator;

// ============================================================================

void DrawButton(int n, int x, int y, char *label, KeySym keycode)
{
    buttonData[n].x1 = x;
    buttonData[n].x2 = x + 16;
    buttonData[n].y1 = y;
    buttonData[n].y2 = y + 16;
    buttonData[n].label = (int) label[0];
    buttonData[n].keycode = keycode;
    XSetForeground(XStuff.display, XStuff.gc, XStuff.fundPixel[0]);
    XDrawRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
		   x, y, 16, 16);
    XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x+4, y+13, 
                     label, strlen(label));
    XSetForeground(XStuff.display, XStuff.gc, textPixel);
}

void Refresh() {

   char *channel_desc[] = {
      "front / back",
      "front face",
      "back face",
      "background",
      "text color",
      "panel color"
   };

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // clear image & z-buffer

   glMatrixMode(GL_MODELVIEW);   // set transformation mode to model/view transformations
   glLoadIdentity();             // clear matrix

   glTranslatef(0.0, 0.0, -(objectRadius + objectDistanceFromCamera));   // move object away from origin, out in front of camera
   glMultMatrixf(Rotator.GetRotationMatrix());   // perform rotations

   sphere.Draw();
   glFlush();
   glXSwapBuffers(XStuff.display, XStuff.window);

   if (bDisplayText) {

      // draw strings
      //
      char buffer[160];
      int x, y, i, j, charHeight, lineSpacing, num;
      GLfloat *rgb;
      
      x = (bRootMode)? 15 : XStuff.windowWidth + 15, 

      XSetForeground(XStuff.display, XStuff.gc, textPixel);
      XSetBackground(XStuff.display, XStuff.gc, panelPixel);

      if (NULL != XStuff.Font) {
         XSetFont(XStuff.display, XStuff.gc, XStuff.Font->fid);
         charHeight = XStuff.Font->ascent + XStuff.Font->descent - 1;
         lineSpacing = (int)(1.6 * charHeight);
         y = 5 + lineSpacing;

         strcpy(buffer,"Press H or ? for key shortcuts");
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         y += lineSpacing;

         i = DATAWIDTH - 26;
         num = 3;
         buttonData[num].x1 = x;
         buttonData[num].x2 = x + i;
         buttonData[num].y1 = y;
         buttonData[num].y2 = y + 17;
         buttonData[num].label = 0;
         buttonData[num].keycode = 0;
         XDrawRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			x, y, i, 17);
         XSetForeground(XStuff.display, XStuff.gc, panelPixel);
         XFillRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			x+1, y+1, i-1, 16);
         XSetForeground(XStuff.display, XStuff.gc, XStuff.fundPixel[0]);
         XFillRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			x+1+(int)( (i-4) * sphere.GetTime()), y+1, 3, 16);
         XSetForeground(XStuff.display, XStuff.gc, textPixel);
         y += 2*lineSpacing;

         sprintf(buffer,"t = %g         ", sphere.GetTime());
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_space);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_space);
         y += lineSpacing;

         sprintf(buffer,"dt = 1/%d  ", (int)(1.0/deltaTime + 0.5));
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_KP_Multiply);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_KP_Divide);
         y += lineSpacing;

         sprintf(buffer,"animation = %s", bAnimatedEversion? "yes  ":"no  "); 
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_F1);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_F1);
         y += lineSpacing;

         sprintf(buffer,"rotation = %s", bAnimatedRotation? "yes  ":"no  ");
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_F2);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_F2);
         y += lineSpacing;

         sprintf(buffer,"zoom = %g             ", frustum.magnificationFactor);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_Up);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_Down);
         y += lineSpacing;

         sprintf(buffer,"number of strips = %d  ", sphere.NumStrips);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_Page_Up);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_Page_Down);
         y += lineSpacing;

         sprintf(buffer,"strips displayed = %d  ", sphere.NumStripsToDisplay);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_Right);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_Left);
         y += lineSpacing;

         sprintf(buffer,"lat precision = %d  ",sphere.NumberOfLatitudinalPatchesPerHemisphere);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_F3);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_F4);
         y += lineSpacing;

         sprintf(buffer,"lon precision = %d  ",sphere.NumberOfLongitudinalPatchesPerStrip);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_F5);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_F6);
         y += lineSpacing;

         sprintf(buffer,"zone selection = S%d H%d", 2-bHalfStrips, sphere.NumHemispheresToDisplay);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_Delete);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_End);
         y += lineSpacing;

         sprintf(buffer,"rendering style = %d", renderingStyle);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_Tab);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_ampersand);
         y += lineSpacing;

         sprintf(buffer,"rotation mode = %d", i = Rotator.GetRotationMode());
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_r);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_R);
         y += lineSpacing;

         switch(i) {
            case RotationInformation::elevationAzimuth : strcpy(buffer,"(elevation-azimuth"); break;
            case RotationInformation::aroundXYAxes     : strcpy(buffer,"(x-y");               break;
            case RotationInformation::virtualSphere    : strcpy(buffer,"(virtual sphere");    break;
            case RotationInformation::arcball          : strcpy(buffer,"(arcball");           break;
            default:
              abort();
              break;
         }
         strcat(buffer, ")                   ");
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         y += 2*lineSpacing;

         sprintf(buffer,"color[%c] = %s        ", '0'+ colorChannel-1, channel_desc[colorChannel]);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         DrawButton(++num, x + DATAWIDTH - 64, y - charHeight, "+", XK_c);
         DrawButton(++num, x + DATAWIDTH - 42, y - charHeight, "-", XK_C);
         y += lineSpacing;

         rgb = &rgbdata[4*(colorChannel-(colorChannel>0))];
         sprintf(buffer,"R = %.3f  G = %.3f  B = %.3f       ",
                 rgb[0], rgb[1], rgb[2]);
         XDrawImageString(XStuff.display, XStuff.mainwindow, XStuff.gc, x, y, buffer, strlen(buffer));
         x += 18;
         y += lineSpacing + 5;

         for (i=0; i<=2; i++) {
 	    buttonData[i].x1 = x + 98*i;
 	    buttonData[i].x2 = x + 98*i + 17;
 	    buttonData[i].y1 = y;
 	    buttonData[i].y2 = XStuff.windowHeight - 20;
            buttonData[i].label = 0;
            buttonData[i].keycode = 0;
            if (buttonData[i].y2 <= buttonData[i].y1 + 32) {
	       // Disable entirely button 
	       buttonData[i].y2 = buttonData[i].y1 - 2;
               continue;
	    }
	    XSetForeground(XStuff.display, XStuff.gc, XStuff.fundPixel[i]);
            XDrawRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			   buttonData[i].x1,  buttonData[i].y1,
                           17, buttonData[i].y2 - buttonData[i].y1);
            j = (int)(rgb[i] *  (buttonData[i].y2 - buttonData[i].y1 - 1));
            XFillRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			   buttonData[i].x1 + 1,  buttonData[i].y1 + 1,
                           16, j);
	    XSetForeground(XStuff.display, XStuff.gc, panelPixel);
            XFillRectangle(XStuff.display, XStuff.mainwindow, XStuff.gc,
			   buttonData[i].x1 + 1,  buttonData[i].y1 + 1 + j,
                           16, buttonData[i].y2 - buttonData[i].y1 - 1 - j);
	 }
      }

      // If the user hits the key to advance time many times in rapid
      // succession, the cpu will start doing some heavy computations
      // to try and keep up, displaying frame after frame as time
      // advances.  Meanwhile, requests to the X server
      // to draw strings can queue up, and finally all get dumped
      // onto the last frame.  The user sees this as text
      // overwriting itself, becoming illegible.  By calling
      // XFlush(), we prevent the requests from queuing up.
      //
      XFlush(XStuff.display);
   }
}

void SetChannelColor(GLfloat *rgb)
{
   switch(colorChannel) {
      case 0:
         glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, rgb);
         rgb[4] = 1.0 - rgb[0];
         rgb[5] = 1.0 - rgb[1];
         rgb[6] = 1.0 - rgb[2];
         glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, &rgb[4]);
         break;
      case 1:
         glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, rgb);
         break;
      case 2:
         glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, rgb);
         break;
      case 3:
         glClearColor(rgb[0], rgb[1], rgb[2], 0.0);
         break;
      case 4:
         FreePixelColor(textPixel);
         textPixel = AllocPixelColor(rgb[0], rgb[1], rgb[2]);
         break;
      case 5:
         FreePixelColor(panelPixel);
         panelPixel = AllocPixelColor(rgb[0], rgb[1], rgb[2]);
         XSetWindowBackground(XStuff.display, XStuff.mainwindow, panelPixel);
         XClearWindow(XStuff.display, XStuff.mainwindow);
         break;
   }
   Refresh();
}

int ButtonCallback(XEvent *event)
{
   GLfloat *rgb;
   double val;
   int i, num, c, x = event->xbutton.x, y = event->xbutton.y;
   num = -1;
   for (i=0; i<=29; i++)
   if (x >= buttonData[i].x1 && x <= buttonData[i].x2 &&
       y >= buttonData[i].y1 && y <= buttonData[i].y2) {
      num = i;
      break;
   }
   if (num >= 0 && num <= 2) {
      // color scrollbar [num] 
      c = colorChannel - (colorChannel>0);
      rgb = &rgbdata[4*c];
      val = (GLfloat)(y - buttonData[num].y1 - 2)/
            ((GLfloat)(buttonData[num].y2 - buttonData[num].y1 - 4));
      if (val < 0.0) val = 0.0;
      if (val > 1.0) val = 1.0;
      rgb[num] = val;
      SetChannelColor(rgb);
      Refresh();
   } else
   if (num == 3) {
      // time scrollbar
      val = (GLfloat)(x - buttonData[num].x1 - 2)/
	    ((GLfloat)(buttonData[num].x2 - buttonData[num].x1 - 3));
      if (val < 0.0) val = 0.0;
      if (val > 1.0) val = 1.0;
      sphere.SetTime(val);
      Refresh();
   }
   // set eversion forward or backward
   if (num == 4) bAnimatedEversionBackwards = False;
   if (num == 5) bAnimatedEversionBackwards = True;
   return num;
}

// ===========================================================================

void MainLoop() {

   Window dummyWindow1, dummyWindow2;
   XEvent event;
   Boolean bQuit = False;
   int newMouseX, newMouseY,
       dummyInt1, dummyInt2;
   KeySym keyCode;
   unsigned int StateOfModifiers;
   int index;
   char buffer[4];
   GLfloat *rgb;

   while (!bQuit) {
      if (!bPause && (bAnimatedEversion || bAnimatedRotation)) {
         while (0 == XEventsQueued(XStuff.display, QueuedAfterFlush)) {
            // Perform animation.
            //
            if (bAnimatedEversion) {
               if (bAnimatedEversionBackwards) {
                  if (sphere.GetTime() == 0.0)
                     bAnimatedEversionBackwards = ! bAnimatedEversionBackwards;
                  else
                     sphere.DecrementTime(deltaTime);
               }
               else {
                  if (sphere.GetTime() == 1.0)
                     bAnimatedEversionBackwards = ! bAnimatedEversionBackwards;
                  else
                     sphere.IncrementTime(deltaTime);
               }
            }
            if (bAnimatedRotation) {
               // Do not rotate if the user is
               // currently trying to rotate.
               //
               if (!Rotator.IsClutchOn()) {
                  double lengthOfMouseDrag;
                  lengthOfMouseDrag = XStuff.minOfWidthAndHeight()*0.5*speedOfAnimatedRotation;
                  Rotator.TurnClutchOn( XStuff.windowWidth/2, XStuff.windowHeight/2 );
                  Rotator.DeviceMotion(
                     XStuff.windowWidth/2  + lengthOfMouseDrag * cos(angleOfMouseDragForAnimatedRotation),
                     XStuff.windowHeight/2 + lengthOfMouseDrag * sin(angleOfMouseDragForAnimatedRotation)
                  );
                  Rotator.TurnClutchOff();
                  angleOfMouseDragForAnimatedRotation += 0.01;
                  if (angleOfMouseDragForAnimatedRotation > 2*PI)
                     angleOfMouseDragForAnimatedRotation -= 2*PI;
               }
            }
            Refresh();
         }
      }

      XNextEvent(XStuff.display, &event);

      switch (event.type) {
         case ClientMessage:
            if (event.xclient.message_type == XStuff.wm_protocols &&
                event.xclient.format == 32 &&
                event.xclient.data.l[0] == XStuff.wm_delete_window)
	       bQuit = True;
            break;
         case Expose:
            // Redirect focus to XStuff.window.
	    if (!bRootMode)
            XSetInputFocus(XStuff.display, XStuff.mainwindow, RevertToPointerRoot, CurrentTime);
            // Unless this is the last contiguous expose, don't redraw.
            if (event.xexpose.count != 0)
               break;
            // Only *contiguous* expose events can be skipped.  Searching through the event queue
            // to find the last expose event would skip over intervening ConfigureNotify events.

            Refresh();
            break;
         case ConfigureNotify:
            // window has been resized
            if (
               (int)XStuff.windowMainWidth != event.xconfigure.width ||
               (int)XStuff.windowHeight != event.xconfigure.height
            ) {
               XStuff.windowMainWidth = event.xconfigure.width;
               XStuff.windowHeight = event.xconfigure.height;
               XStuff.windowWidth = XStuff.windowMainWidth - DATAWIDTH;
               XClearWindow(XStuff.display, XStuff.mainwindow);
               XResizeWindow(XStuff.display, XStuff.window,
                             XStuff.windowWidth, XStuff.windowHeight);
               frustum.WindowResize();
               Rotator.CalibrateDevice(XStuff.windowWidth/2, XStuff.windowHeight/2, XStuff.minOfWidthAndHeight()/2);
               Refresh();
            }
            break;
         case KeyPress:

            // Consult /usr/include/X11/keysymdef.h for keysym codes
            //
           XLookupString(&event.xkey, buffer, 1, &keyCode, NULL);
      do_click_action:
	   switch (keyCode) {
               case XK_KP_Add:
               // case XK_plus:
                  sphere.IncrementTime(deltaTime);
                  Refresh();
                  break;
               case XK_KP_Subtract:
               // case XK_minus:
                  sphere.DecrementTime(deltaTime);
                  Refresh();
                  break;
               case XK_KP_Multiply:
                  deltaTime *= 2.0;
                  if (deltaTime > 1.0)
                     deltaTime = 1.0;
                  Refresh();
                  break;
               case XK_KP_Divide:
                  if (deltaTime * 0.5 >= 1.0/8192)
                     deltaTime *= 0.5;
                  Refresh();
                  break;
               case XK_Page_Up:
                  sphere.IncrementNumStrips();
                  Refresh();
                  break;
               case XK_Page_Down:
                  sphere.DecrementNumStrips();
                  Refresh();
                  break;
               case XK_Up:
                  frustum.Magnify(1);
                  Refresh();
                  break;
               case XK_Down:
                  frustum.Magnify(-1);
                  Refresh();
                  break;
               case XK_Home:
                  frustum.Reset();
                  Refresh();
                  break;
               case XK_Right:
                  sphere.IncrementNumStripsToDisplay();
                  Refresh();
                  break;
               case XK_Left:
                  sphere.DecrementNumStripsToDisplay();
                  Refresh();
                  break;
               case XK_End:
                  sphere.ChangeNumHemispheresToDisplay();
                  Refresh();
                  break;
               case XK_Delete:
                  bHalfStrips = !bHalfStrips;
                  sphere.Reconstruct();
                  Refresh();
                  break;
               case XK_space:
                  if (!bAnimatedEversion && !bAnimatedRotation) {
		     bPause = False;
		     bAnimatedEversion = bAnimatedRotation = True;
		  } else
		     bPause = !bPause;
		  Refresh();
                  break;
               case XK_F1:
                  bAnimatedEversion = !bAnimatedEversion;
                  Refresh();
                  break;
               case XK_F2:
                  bAnimatedRotation = !bAnimatedRotation;
                  Refresh();
                  break;
               case XK_F3:
                  sphere.IncrementLatitudinalResolution();
                  Refresh();
                  break;
               case XK_F4:
                  sphere.DecrementLatitudinalResolution();
                  Refresh();
                  break;
               case XK_F5:
                  sphere.IncrementLongitudinalResolution();
                  Refresh();
                  break;
               case XK_F6:
                  sphere.DecrementLongitudinalResolution();
                  Refresh();
                  break;
               case XK_F7:
		  sphere.SetDefaultResolution();
                  Refresh();
                  break;
               case XK_F8:
                  bDisplayText = ! bDisplayText;
                  if (bDisplayText) {
		     XStuff.windowWidth = XStuff.windowMainWidth - DATAWIDTH;
		     XResizeWindow(XStuff.display, XStuff.window, 
                                   XStuff.windowWidth, XStuff.windowHeight);
		     XClearWindow(XStuff.display, XStuff.mainwindow);
                     frustum.WindowResize();
		  } else {
		     XStuff.windowWidth = XStuff.windowMainWidth;
		     XResizeWindow(XStuff.display, XStuff.window, 
                                   XStuff.windowWidth, XStuff.windowHeight);
		     XClearWindow(XStuff.display, XStuff.mainwindow);
                     frustum.WindowResize();
		  }
                  break;
#ifdef TOGGLEABLE_WINDOW_DECORATIONS
               case XK_F10:
                  PropMotifWmHints mwm_hints;
                  Atom mwm_hints_atom;

                  bNoWindowDecorations = ! bNoWindowDecorations;
                  mwm_hints_atom  = XInternAtom(
                     XStuff.display, _XA_MOTIF_WM_HINTS, False
                  );
                  mwm_hints.flags = (int) MWM_HINTS_DECORATIONS;
                  mwm_hints.decorations = bNoWindowDecorations ? 0
                     : MWM_DECOR_ALL;
                     //: MWM_DECOR_BORDER | MWM_DECOR_TITLE | MWM_DECOR_RESIZEH
                     //| MWM_DECOR_MINIMIZE | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU;
                  XChangeProperty (
                     XStuff.display, XStuff.window, mwm_hints_atom,
                     mwm_hints_atom,
                     32, PropModeReplace, (unsigned char *) & mwm_hints,
                     PROP_MWM_HINTS_ELEMENTS
                  );
                  break;
#endif
               case XK_Tab:
                  renderingStyle = (RenderingStyle)(((int)renderingStyle + 1) % number_of_styles);
                  Refresh();
                  break;
               case XK_ampersand:
                  renderingStyle = (RenderingStyle)(((int)renderingStyle + number_of_styles - 1) % number_of_styles);
                  Refresh();
                  break;
               case XK_s:
               case XK_S:
                  bShadingIsSmooth = !bShadingIsSmooth;
                  glShadeModel(bShadingIsSmooth ? GL_SMOOTH : GL_FLAT);
                  Refresh();
                  break;
               case XK_b:
               case XK_B:
                  // toggle back face culling
                  bBackFaceCulling = !bBackFaceCulling;
                  if (bBackFaceCulling) {
                     glEnable(GL_CULL_FACE);
                     glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
                  }
                  else
                     glDisable(GL_CULL_FACE);
                  Refresh();
                  break;
               case XK_f:
               case XK_F:
                  bFrontFacing = !bFrontFacing;
                  glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
                  Refresh();
                  break;
               case XK_r:
                  Rotator.CycleRotationMode(1);
                  Refresh();
                  break;
               case XK_R:
                  // cycle forward to new rotation mode
                  Rotator.CycleRotationMode(5);
                  Refresh();
                  break;
	       case XK_c:
		  colorChannel = (colorChannel + 1) % 6;
                  Refresh();
                  break;
	       case XK_C:
		  colorChannel = (colorChannel + 5) % 6;
                  Refresh();
                  break;
               case XK_0:
               case XK_1:
               case XK_2:
               case XK_3:
               case XK_4:
               case XK_5:
               case XK_6:
               case XK_7:
               case XK_8:
               case XK_9:
                  index = keyCode - XK_1;
                  goto set_values;
               case XK_KP_0:
               case XK_KP_1:
               case XK_KP_2:
               case XK_KP_3:
               case XK_KP_4:
               case XK_KP_5:
               case XK_KP_6:
               case XK_KP_7:
               case XK_KP_8:
               case XK_KP_9:
                  index = keyCode - XK_KP_1;
	    set_values:
                  // change colour of something
                  rgb = &rgbdata[4*(colorChannel-(colorChannel>0))];

                  if (index<0) {
                     int i;
		     for (i=0; i<=2; i++) rgb[i] = 0.96 * rgb[i];
		  } else
                  if (index>7) {
                     int i;
		     for (i=0; i<=2; i++) {
                        rgb[i] = rgb[i] / 0.96;
                        if (rgb[i] > 1.0) rgb[i] =  1.0;
		     }
		  } else {
                     rgb[0] = (GLfloat)(index & 1);
                     rgb[1] = (GLfloat)((index >> 1) & 1);
                     rgb[2] = (GLfloat)((index >> 2) & 1);
		  }
                  SetChannelColor(rgb);
                  break;
               case XK_h:
               case XK_H:
               case XK_question:
  		  bPause = True;
		  Help();
                  break;
            }
            break;
         case KeyRelease:
            break;
         case ButtonRelease:
         case ButtonPress:
	    if (event.type == ButtonRelease) {
               Rotator.TurnClutchOff();
	       break;
	    }
	    if (event.xbutton.window == XStuff.mainwindow) {
	       index = ButtonCallback(&event);
               if (index >= 4) {
		  keyCode = buttonData[index].keycode;
                  if (bDisplayText && keyCode) goto do_click_action;
                  break;
	       }
	    }
            XQueryPointer(XStuff.display, XStuff.window, &dummyWindow1,
                        &dummyWindow2, &dummyInt1, &dummyInt2,
                        &newMouseX, &newMouseY, &StateOfModifiers);
#ifndef TRY_SOMETHING_WEIRD
            Rotator.TurnClutchOn(newMouseX, newMouseY);
#else
            Rotator.TurnClutchOn(windowWidth/2, windowHeight/2);
            Rotator.DeviceMotion(newMouseX,newMouseY);
            Refresh();
#endif
            break;
         case MotionNotify:
            if (Rotator.IsClutchOn()) {
               XQueryPointer(XStuff.display, XStuff.window, &dummyWindow1,
                        &dummyWindow2, &dummyInt1, &dummyInt2,
                        &newMouseX, &newMouseY, &StateOfModifiers);
               Rotator.DeviceMotion(newMouseX,newMouseY);
               Refresh();
            }
            break;
      }   /* switch(event */
   }   /* while (!bQuit) */
}

// ==============================================================================

void Usage()
{
   fprintf(stderr,"Usage: sphereEversion [--root] [--command]\n");
   exit(1);
}

int main(int argc, char *argv[]) {

   char * display_name = NULL;       /* Server to connect to */
   int i;
   XColor xc;

   XStuff.commandLineName = basename(argv[0]);

   for (i=1; i<argc; i++) {
      if (!strcmp(argv[i], "--root")) {
         bRootMode = True;
         bPause = False;
      } else
      if (!strcmp(argv[i], "--command")) {
         bCmdMode = True;
      } else
         Usage();
   }

   // Connect to the X server.
   //
   if (NULL == (XStuff.display = XOpenDisplay(display_name))) {
      fprintf(stderr,"%s: can't connect to X server %s\n",
         XStuff.commandLineName, XDisplayName(display_name));
      exit(1);
   }

   XStuff.wm_protocols = XInternAtom(XStuff.display, "WM_PROTOCOLS", False);
   XStuff.wm_delete_window = XInternAtom(XStuff.display, "WM_DELETE_WINDOW", False);

   XStuff.screen_num     = DefaultScreen         (XStuff.display);
   XStuff.screen_ptr     = DefaultScreenOfDisplay(XStuff.display);
   XStuff.screenWidth    = DisplayWidth          (XStuff.display, XStuff.screen_num);
   XStuff.screenHeight   = DisplayHeight         (XStuff.display, XStuff.screen_num);
   XStuff.screenWidthMM  = DisplayWidthMM        (XStuff.display, XStuff.screen_num);
   XStuff.screenHeightMM = DisplayHeightMM       (XStuff.display, XStuff.screen_num);
   XStuff.rootWindow     = RootWindow            (XStuff.display, XStuff.screen_num);
   XStuff.blackPixel     = BlackPixel            (XStuff.display, XStuff.screen_num);
   XStuff.whitePixel     = WhitePixel            (XStuff.display, XStuff.screen_num);

   MakeWindow("Sphere Eversion","Eversion",None,0,0,True,XStuff.blackPixel,argc,argv);

   if (XAllocNamedColor(XStuff.display, XStuff.colorMap, "red", &xc, &xc))
      XStuff.fundPixel[0] = xc.pixel;
   else
      XStuff.fundPixel[0] = XStuff.whitePixel;
   if (XAllocNamedColor(XStuff.display, XStuff.colorMap, "green", &xc, &xc))
      XStuff.fundPixel[1] = xc.pixel;
   else
      XStuff.fundPixel[1] = XStuff.whitePixel;
   if (XAllocNamedColor(XStuff.display, XStuff.colorMap, "blue", &xc, &xc))
      XStuff.fundPixel[2] = xc.pixel;
   else
      XStuff.fundPixel[2] = XStuff.whitePixel;

   if (!bRootMode ) 
      CreateCursorForWindow();
   else
      XStuff.cursor = None;

   // Initialize some OpenGL stuff.

   if (bBackFaceCulling) {
      // turn on back face culling
      glEnable(GL_CULL_FACE);
      glCullFace(bFrontFacing ? GL_BACK : GL_FRONT);
   }

   // Don't bother doing this, since the normals are
   // already normalized by us at generation time.
   // There's no need to ask OpenGL to explicitly try
   // to normalize again and again every time the
   // user rotates the view.
   //
   // glEnable(GL_NORMALIZE);   // This asks that normal vectors be normalized for us, automatically

   glClearDepth(1.0);
   // glDepthFunc(GL_LEQUAL);
   glEnable(GL_DEPTH_TEST);

   // Don't bother doing this
   //
   // // Clear the buffers.
   // //
   // glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // glXSwapBuffers(display, window);
   // glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   // glXSwapBuffers(display, window);

   // Setup colors & lighting properties of the materiel.
   // rgbdata: 4 GLfloats per channel
   // channels 0:front 1:back 2:background 3:text 4:panel 5:specular
    
   // alpha levels
   for (i=0; i<6; i++) rgbdata[3+4*i] = 1.0;

   // color of front faces of object
   rgbdata[0] = 0.0; rgbdata[1] = 1.0; rgbdata[2] = 1.0;
   glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, &rgbdata[0]);
   // color of back faces of object
   rgbdata[4] = 1.0; rgbdata[5] = 0.0; rgbdata[6] = 0.0; 
   glMaterialfv(GL_BACK, GL_AMBIENT_AND_DIFFUSE, &rgbdata[4]);

   // this sets the colour to use when glClear() is called
   glClearColor(rgbdata[8]=0.0, rgbdata[9]=0.0, rgbdata[10]=0.0, 0.0);

   // text color
   textPixel = AllocPixelColor(rgbdata[12]=1.0, rgbdata[13]=1.0, rgbdata[14]=1.0);
   // panel color
   panelPixel = AllocPixelColor(rgbdata[16]=0.0432, rgbdata[17]=0.0863, rgbdata[18]=0.1295);

   XSetForeground(XStuff.display, XStuff.gc, textPixel);
   XSetWindowBackground(XStuff.display, XStuff.mainwindow, panelPixel);
   XClearWindow(XStuff.display, XStuff.mainwindow);

   // specular color
   rgbdata[20] = 1.0; rgbdata[21] = 1.0; rgbdata[22] = 1.0; 
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, &rgbdata[20]);
   GLfloat shininess = 50.0;
   glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &shininess);

   // Setup lighting.
   //
   GLfloat lightPosition[] = { 10.0, 10.0, 10.0, 0.0 }; 
   glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   GLfloat lightModelFlag[1] = { 1.0 };
   glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lightModelFlag);

   // Initialze some of my own stuff.
   //
   Rotator.CalibrateDevice(XStuff.windowWidth/2,XStuff.windowHeight/2,XStuff.minOfWidthAndHeight()/2);
   frustum.Init();

   // Loop for events until user quits.
   //
   MainLoop();

   // Clean up.
   //
   FreePixelColor(textPixel);
   FreePixelColor(panelPixel);
   if (NULL != XStuff.Font)
      XUnloadFont(XStuff.display, XStuff.Font->fid);
   XFreeGC(XStuff.display, XStuff.gc);
   if (XStuff.cursor)
      XFreeCursor(XStuff.display, XStuff.cursor);
   XCloseDisplay(XStuff.display);
   return 0;
}

